ImplicitReplyActivity subclass: #ExpressImplicitReplyActivity
	instanceVariableNames: 'expressProcess expressExclusionSemaphore '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-ExpressSend'!
ExpressImplicitReplyActivity comment:
'Class ExpressAbclActorBehavior computes express mode messages.
A second behavior autonomous process is created for handling express messages. When resumed (when an express message arrived), it suspends standard behavior process while computing an express message.
Method nonResume may be used within an express message method. It will abort resumption of computing a normal message if one was being interrupted.
Method expressAtomic: may be used within any standard method. It ensures that computation won''t be interrupted by a possibly incoming express message.

Class ExpressAbclActorBehavior is a subclass of class AbclActorBehavior.

Instance Variables:

	expressProcess				<Process>		the process for computing express messages.
	expressExclusionSemaphore 	<Semaphore>	ensures mutual exclusion of express messages
												with the expressAtomic: construct.'!


!ExpressImplicitReplyActivity methodsFor: 'initialize'!

privateInitialize
	"Initialize the semaphore for ensuring atomicity against express message interruption."

	super privateInitialize.
	expressExclusionSemaphore := Semaphore forMutualExclusion! !

!ExpressImplicitReplyActivity methodsFor: 'accessing'!

expressExclusionSemaphore
	^expressExclusionSemaphore! !

!ExpressImplicitReplyActivity methodsFor: 'activity setting'!

startProcess
	"Create and start another activity process specific for express messages."

	super startProcess.
	expressProcess :=
		[[true] whileTrue:
			[self acceptExpressMessage: address expressMailBox next]] newProcess.
	expressProcess resume! !

!ExpressImplicitReplyActivity methodsFor: 'process control'!

terminate
	"Terminate also express message behavior process."

	"We have to launch a specific process to terminate
	both behavior process and express behavior process.
	At first look it may seem very paradoxal indeed!!
	However the point is that we don't know which current process is:
	behavior process, express behavior process, or an external process.
	We have to terminate two processes in sequence.
	If the first process we terminate happens to be current process
	we would never terminate the second one!!"

	[super terminate.
	expressProcess terminateNoSignal] fork! !

!ExpressImplicitReplyActivity methodsFor: 'message handling'!

acceptExpressMessage: aMessage
	"Accepting an express message takes priority (suspend) over normal (standard) acceptance.
	Mutually exclude with the expressAtomic: construct."

	expressExclusionSemaphore critical:
		[process suspend.
		self acceptMessage: aMessage.
		process resume]! !

!ExpressImplicitReplyActivity methodsFor: 'default classes'!

addressClass
	^ExpressImplicitReplyAddress! !

!ExpressImplicitReplyActivity methodsFor: 'compatibility constraints'!

addressConstraint
	^ExpressImplicitReplyAddress! !

ImplicitReplyAddress subclass: #ExpressImplicitReplyAddress
	instanceVariableNames: 'expressMailBox '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-ExpressSend'!


!ExpressImplicitReplyAddress methodsFor: 'initialize'!

privateInitialize
	"Initialize the express mailbox."

	super privateInitialize.
	expressMailBox := MailBox new! !

!ExpressImplicitReplyAddress methodsFor: 'accessing'!

expressMailBox
	^expressMailBox! !

!ExpressImplicitReplyAddress methodsFor: 'message passing'!

receiveMessage: aMessage withSendType: type withSendMode: mode
	"Dispatch along the express mailBox according to the mode."

	mode == #express
		ifTrue: [^self receiveMessage: aMessage withSendType: type inMessageQueue: expressMailBox].
	^super receiveMessage: aMessage withSendType: type withSendMode: mode!

sendModeOf: aMessage
	"An express mode message is tagged through the first letter of the selector (E for express)."

	^(aMessage selector at: 1) = $E
		ifTrue: [#express]
		ifFalse: [#standard]! !

ImplicitReplyObject subclass: #ExpressImplicitReplyObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-ExpressSend'!


!ExpressImplicitReplyObject methodsFor: 'express method construct'!

expressAtomic: aBlock
	"Ensure that computation expressed within aBlock won't be interrupted by an express message.
	May be used within script methods as well as private routines."
	
	"WARNING: This construct should NOT be used within an express method,
	otherwise the express method computation will DEADLOCK!!!!"

	self activity expressExclusionSemaphore critical: aBlock! !

!ExpressImplicitReplyObject methodsFor: 'express control construct'!

nonResume
	"Terminate current standard behavior process and recreate a new one.
	In case a normal message was being processed, it will be aborted."

	"Should be used ONLY by express message methods!!!!
	(Otherwise if used within standard methods,
	it would abort main behavior process which is current process!!!!)."

	"Note: the new process will be only created and not started,
	so that it will be in the suspend state as was current process to be resumed."

	self activity process terminateNoSignal.
	self activity createProcess! !

!ExpressImplicitReplyObject methodsFor: 'default classes'!

activityClass
	^ExpressImplicitReplyActivity! !

!ExpressImplicitReplyObject methodsFor: 'compatibility constraints'!

activityConstraint
	^ExpressImplicitReplyActivity! !
