Basic2Address subclass: #Address
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel'!
Address comment:
'Classes Actor and ActorBehavior are the two kernel classes, entry point of Actalk system.

Class Actor is the class which implements the external shell of an actor.
It acts mostly as a container for the mailbox which will contain incoming messages and the behavior which will process them.

See class ActorBehavior comment and examples in the Actalk-Examples category.'!


!Address methodsFor: 'compatibility constraints'!

activityConstraint
	"Default bottom compatibility constraint."

	^Activity!

objectConstraint
	"Default bottom compatibility constraint."

	^ActiveObject! !

Basic2ActiveObject subclass: #ActiveObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel'!
ActiveObject comment:
'Classes Actor and ActorBehavior are the two kernel classes, entry point of Actalk system.
They are just dedicated class names and are subclasses of basic kernel classes where the semantics is defined. (Classes Basic1Actor and Basic1ActorBehavior are the main actual implementation classes of the kernel.)

Class ActorBehavior implements the behavior of an actor.
It is the object which eventually consumes incoming messages and processes them.
Kernel class actors compute messages serially. There is inter-object concurrency (between various object activities) but no intra-object concurrency, that is at most one activity with the actor. (See classes ReplaceActorBehavior and CountersSynchroActorBehavior as examples of inter-object concurrency).
Message passing is asynchronous, that is the sender does not wait for completion of the computation of its request by the receiver. If the receiver returns a reply it should send the returned value through another independent message transmission (default reply selector is by convention the selector reply:, see below, as well as explanation in class BlockContinuation comment, and examples in category Actalk-Examples). Classes like AbclActor, IRActor and others provide others types of message passing, like synchronous and also future type (with an eager reply).

Class ActorBehavior implements the following current features:

Class ActorBehavior provides the way to create (anonymous) block continuation actors as a more concise alternative to explicit classes of continuations. (See their definition in class BlockContinuation and example of use within class Factorial.)
Class ActorBehavior also provides a way to reply to a reply destination with a (computed) reply selector passed as an argument along with the reply destination. (Note: this has not been necessary up to now, but this may prove useful when connecting various independent preexisting modules, where reply selectors may be different, as example selector store: to store a value). See example in class Counter within category Actalk-Examples.

As for class methods, class ActorBehavior provides two important and useful facilities.
Tracing facilities automatically set (or unset) traces onto a specific class by adding (or removing) user event methods which generate trace information onto the transcript window.
Cleanup facilities provide a termination of actor behavior processes of a specific class or some sub hierarchy. This proves to be necessary to help the garbage collector (gc) recovering Actalk processes.

For some general information about Actalk, its objectives, architecture, and way of using, please read Actalk info file INFO.
Following are some sections about classes Actor and ActorBehavior design and use:

The pair of classes Actor and ActorBehavior model the notion of active object (that is which owns its activity) and which communicates with other active objects through message passing, possibly asynchronous (without waiting for the reply). We call such active objects by the generic name of actors.
In general, Actalk is aimed at modeling various kind of actor and also called object-oriented concurrent programming models and languages (that is various models of object activity, message passing, synchronization...) as proposed by various programming languages of that kind. (See the book OOCP, edited by Akinori Yonezawa and Mario Tokoro, MIT-Press, 1987, for an introduction. Actalk extensions model and implement most of constructs and models described within this book). These modelings are implemented as subclasses of the two kernel classes. Therefore Actalk provides a testbed environment where various actor programming models and constructs are implemented and may be related and compared.

Note that the kernel is decomposed into two kernel classes: Actor and ActorBehavior to give maximum genericity to modeling abilities. For instance in some models with intra-object concurrency, there could be several behaviors active within the same actor. Meanwhile this decomposition imposes minimal consistency between classes of Actor kind and classes of ActorBehavior kind. For instance subclasses IRActor and IRActorBehavior, which implement implicit reply, should be paired, see class IRActorBehavior comment.

To define an Actalk actor, there are four differences compared with standard Smalltalk programming.
First, the class defining behaviors should be defined as a subclass of class ActorBehavior and not as a subclass of class Object.
Second, when creating an instance, message actor should be sent to this new behavior in order to construct the final actor.
Third, because message passing is asynchronous, no implicit reply is returned as a value of the message transmission. If a reply is needed, one should explicitly program the reply message, as well as conveying the reply destination within the initial message. See category Actalk-Examples as well as category Actalk-Kernel-JoinCont for predefined reply destination behaviors. (Note also that classes IRActor and IRActorBehavior provide implicit reply as in standard Smalltalk-80).
Fourth, the self reference of an actor is pseudo variable aself (and not self). Seen from within the behavior (context), aself represents the actor of which it is a behavior. self represents the behavior itself. By sending a message to aself, the actor behavior sends (asynchronously through its mailbox) a message to itself (as an actor). By sending a message to self, the actor behavior sends (as in standard Smalltalk) a message directly to itself (behavior).

To clearly distinguish within a behavior between standard Smalltalk methods (for instance to initialize it), and message patterns intended to be sent to the actor, by convention we group the latter set within the category protocol named script. Script methods are intended to be requested through actor message passing.
Alternatively, when modeling medium grained or large grained actors, inner computation is described as standard Smalltalk message passing between standard object sub components of the behavior. By convention such private inner computation methods are grouped within category protocol named private routines (by using ABCL/1 terminology). They can only be requested by the behavior itself by sending messages to self.

Please start with examples in the Actalk-Examples category.'!


!ActiveObject methodsFor: 'replying'!

reply: value withSelector: replySelector
	"Reply with a specific reply selector which needs to be computed.
	Basically it has been passed as an argument. See example in class Counter."

	self perform: replySelector with: value! !

!ActiveObject methodsFor: 'continuation creation'!

continuationBlock: aBlock
	"Create a block continuation actor with the block (closure) as a behavior.
	The block has two arguments: the replied value, and the actor continuation actor (usually named me)."
	"See class BlockContinuation."

	^(BlockContinuation new behaviorBlock: aBlock) active!

singleReplyContinuationBlock: aBlock
	"Create a block continuation actor with the block (closure) as a behavior.
	The block has two arguments: the replied value, and the actor continuation actor (usually named me)."
	"See class SingleReplyBlockContinuation."

	^(SingleReplyBlockContinuation new behaviorBlock: aBlock) active! !

!ActiveObject methodsFor: 'default classes'!

activityClass
	^Activity!

addressClass
	^Address! !

!ActiveObject methodsFor: 'compatibility constraints'!

activityConstraint
	"Default bottom compatibility constraint."

	^self activityClass!

addressConstraint
	"Default bottom compatibility constraint."

	^self addressClass! !

!ActiveObject methodsFor: 'compatibility'!

actor
	"Compatibility with version 3.02."

	"While checking old version scories."
	self error: 'still using actor'.

	^self active! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ActiveObject class
	instanceVariableNames: ''!


!ActiveObject class methodsFor: 'event tracing'!

setTrace: setOrUnset
	"Set (or unset) tracing facilities for this class onto all three events.
	(aClass setTrace: true) to set, (aClass setTrace: false) to unset.
	It does not check whenever some event methods have already been defined.
	It simply adds them or removes them."

	self setTraceOnReceive: setOrUnset;
		setTraceOnAccept: setOrUnset;
		setTraceOnComplete: setOrUnset!

setTraceOnAccept: setOrUnset
	"Set (or unset) tracing facility for this class onto the event accept.
	It does not check whenever an event methods has already been defined.
	It simply adds it or removes it."

	setOrUnset
		ifTrue: [self compile:
'eventAccept: aMessage
	Transcript show: self printString , '' accept ('' , aMessage compactPrintString , '')''; cr'
				classified: #'events']
		ifFalse: [self removeSelector: #eventAccept:]!

setTraceOnComplete: setOrUnset
	"Set (or unset) tracing facility for this class onto the complete accept.
	It does not check whenever an event methods has already been defined.
	It simply adds it or removes it."

	setOrUnset
		ifTrue: [self compile:
'eventComplete: aMessage
	Transcript show: self printString , '' complete ('' , aMessage compactPrintString , '')''; cr'
				classified: #'events']
		ifFalse: [self removeSelector: #eventComplete:]!

setTraceOnReceive: setOrUnset
	"Set (or unset) tracing facility for this class onto the event receive.
	It does not check whenever an event methods has already been defined.
	It simply adds it or removes it."

	setOrUnset
		ifTrue: [self compile:
'eventReceive: aMessage
	Transcript show: self printString , '' receive ('' , aMessage compactPrintString , '')''; cr'
				classified: #'events']
		ifFalse: [self removeSelector: #eventReceive:]!

unsetAllTraces
	"Remove ALL event tracing methods in the Actalk hierarchy.
	Useful for doing some cleanup like before filing out."

	ActiveObject withAllSubclasses do:
		[:aClass | aClass setTrace: false]! !

!ActiveObject class methodsFor: 'process cleanup'!

allCleanUp
	"Provide a reset/cleanup of activity processes of that kind.
	Delegate allCleanUp to the associated activity class."

	"ActiveObject allCleanUp"

	| instance |						"To find the associated activity class,"
	instance := self someInstance.		"we need one of its instances,"
	instance isNil					"possibly by creating one."
		ifTrue: [instance := self basicNew].
	instance activityClass allCleanUp!

cleanUp
	"Provide a reset/cleanup of activity processes associated to active objects of this class.
	Delegate cleanUp to the associated activity class."
	"See method allCleanUp to also cleanup all subclasses."

	| instance |						"To find the associated activity class,"
	instance := self someInstance.		"we need one of its instances,"
	instance isNil					"possibly by creating one."
		ifTrue: [instance := self new].
	instance activityClass cleanUp! !

!ActiveObject class methodsFor: 'compatibility constraints'!

allCheckConstraints
	"Check constraints for all active object classes."

	"ActiveObject allCheckConstraints"

	| isSatisfied |
	isSatisfied := true.
	Transcript show: 'Constraints check for ' , self name , ' hierarchy start'.
	self checkConstraints.
	self allSubclasses do: [:anActiveObjectClass |
		isSatisfied := isSatisfied & (anActiveObjectClass checkConstraints).
		Transcript show: '.'].
	Transcript show: ' done!!'; cr.
	isSatisfied
		ifTrue: [Transcript show: 'All classes of ' , self name , ' hierarchy are compatible.'; cr]
		ifFalse: [Transcript show: 'Some classes of ' , self name , ' hierarchy are incompatible.';
						cr;
 						show: 'Please check the Transcript window.'].
	^isSatisfied!

checkConstraint: class1 with: class2
	"Check one compatibility constraint by testing if class1 is equal or a subclass of class2.
	(class2 may be an ordered collection of classes.)
	Return the boolean result of the check."

	| isSatisfied |
	isSatisfied :=
		class2 isBehavior
			ifTrue:
				["Only one class as a constraint."
				class1 isEqualOrSubclassOf: class2]
			ifFalse:
				[class2 isSequenceable
					ifFalse:
						["The constraint is neither a class nor an ordered collection of classes."
						self error: 'compatibility constraint not valid']
					ifTrue:
						["The constraint is an ordered collection of classes.
						At least one of them should be verified."
						class2 inject: false into: [:boolean :class |
												boolean | (class1 isEqualOrSubclassOf: class)]]].
	isSatisfied ifFalse:
		[Transcript show: 'constraint compatibility violation : '
						, class1 name , ' should be a subclass of ' , class2 printString; cr].
	^isSatisfied!

checkConstraints
	"Check if active object class, activity class, and addressclass are mutually compatible."

	| isSatisfied object activity address |
	"We need to create an instance, and its associated activity and address."
	"Note that we are using basicNew in order to bypass any initialization method."
	object := self basicNew.					
	activity := object activityClass basicNew.
	address := (activity computeAddressClass: object addressClass) basicNew.
	isSatisfied :=		"Conjonction of all 6 constraint checks."
	"Check constraints specified by the active object class."
		(self checkConstraint: activity class with: object activityConstraint)
		& (self checkConstraint: address class with: object addressConstraint)
	"Check constraints specified by the activity class."
		& (self checkConstraint: object class with: activity objectConstraint)
		& (self checkConstraint: address class with: activity addressConstraint)
	"Check constraints specified by the address class."
		& (self checkConstraint: object class with: address objectConstraint)
		& (self checkConstraint: activity class with: address activityConstraint).
	isSatisfied ifFalse:
		[Transcript show: self name
				, ' has compatibility problem with: '
				, activity class name , ' and ' , address class name; cr].
	^isSatisfied! !

Basic2Activity subclass: #Activity
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel'!
Activity comment:
'Classes Actor and ActorBehavior are the two kernel classes, entry point of Actalk system.
They are just dedicated class names and are subclasses of basic kernel classes where the semantics is defined. (Classes Basic1Actor and Basic1ActorBehavior are the main actual implementation classes of the kernel.)

Class ActorBehavior implements the behavior of an actor.
It is the object which eventually consumes incoming messages and processes them.
Kernel class actors compute messages serially. There is inter-object concurrency (between various object activities) but no intra-object concurrency, that is at most one activity with the actor. (See classes ReplaceActorBehavior and CountersSynchroActorBehavior as examples of inter-object concurrency).
Message passing is asynchronous, that is the sender does not wait for completion of the computation of its request by the receiver. If the receiver returns a reply it should send the returned value through another independent message transmission (default reply selector is by convention the selector reply:, see below, as well as explanation in class BlockContinuation comment, and examples in category Actalk-Examples). Classes like AbclActor, IRActor and others provide others types of message passing, like synchronous and also future type (with an eager reply).

Class ActorBehavior implements the following current features:

Class ActorBehavior provides the way to create (anonymous) block continuation actors as a more concise alternative to explicit classes of continuations. (See their definition in class BlockContinuation and example of use within class Factorial.)
Class ActorBehavior also provides a way to reply to a reply destination with a (computed) reply selector passed as an argument along with the reply destination. (Note: this has not been necessary up to now, but this may prove useful when connecting various independent preexisting modules, where reply selectors may be different, as example selector store: to store a value). See example in class Counter within category Actalk-Examples.

As for class methods, class ActorBehavior provides two important and useful facilities.
Tracing facilities automatically set (or unset) traces onto a specific class by adding (or removing) user event methods which generate trace information onto the transcript window.
Cleanup facilities provide a termination of actor behavior processes of a specific class or some sub hierarchy. This proves to be necessary to help the garbage collector (gc) recovering Actalk processes.

For some general information about Actalk, its objectives, architecture, and way of using, please read Actalk info file INFO.
Following are some sections about classes Actor and ActorBehavior design and use:

The pair of classes Actor and ActorBehavior model the notion of active object (that is which owns its activity) and which communicates with other active objects through message passing, possibly asynchronous (without waiting for the reply). We call such active objects by the generic name of actors.
In general, Actalk is aimed at modeling various kind of actor and also called object-oriented concurrent programming models and languages (that is various models of object activity, message passing, synchronization...) as proposed by various programming languages of that kind. (See the book OOCP, edited by Akinori Yonezawa and Mario Tokoro, MIT-Press, 1987, for an introduction. Actalk extensions model and implement most of constructs and models described within this book). These modelings are implemented as subclasses of the two kernel classes. Therefore Actalk provides a testbed environment where various actor programming models and constructs are implemented and may be related and compared.

Note that the kernel is decomposed into two kernel classes: Actor and ActorBehavior to give maximum genericity to modeling abilities. For instance in some models with intra-object concurrency, there could be several behaviors active within the same actor. Meanwhile this decomposition imposes minimal consistency between classes of Actor kind and classes of ActorBehavior kind. For instance subclasses IRActor and IRActorBehavior, which implement implicit reply, should be paired, see class IRActorBehavior comment.

To define an Actalk actor, there are four differences compared with standard Smalltalk programming.
First, the class defining behaviors should be defined as a subclass of class ActorBehavior and not as a subclass of class Object.
Second, when creating an instance, message actor should be sent to this new behavior in order to construct the final actor.
Third, because message passing is asynchronous, no implicit reply is returned as a value of the message transmission. If a reply is needed, one should explicitly program the reply message, as well as conveying the reply destination within the initial message. See category Actalk-Examples as well as category Actalk-Kernel-JoinCont for predefined reply destination behaviors. (Note also that classes IRActor and IRActorBehavior provide implicit reply as in standard Smalltalk-80).
Fourth, the self reference of an actor is pseudo variable aself (and not self). Seen from within the behavior (context), aself represents the actor of which it is a behavior. self represents the behavior itself. By sending a message to aself, the actor behavior sends (asynchronously through its mailbox) a message to itself (as an actor). By sending a message to self, the actor behavior sends (as in standard Smalltalk) a message directly to itself (behavior).

To clearly distinguish within a behavior between standard Smalltalk methods (for instance to initialize it), and message patterns intended to be sent to the actor, by convention we group the latter set within the category protocol named script. Script methods are intended to be requested through actor message passing.
Alternatively, when modeling medium grained or large grained actors, inner computation is described as standard Smalltalk message passing between standard object sub components of the behavior. By convention such private inner computation methods are grouped within category protocol named private routines (by using ABCL/1 terminology). They can only be requested by the behavior itself by sending messages to self.

Please start with examples in the Actalk-Examples category.'!


!Activity methodsFor: 'default classes'!

addressClass
	^Address! !

!Activity methodsFor: 'compatibility constraints'!

addressConstraint
	"Default bottom compatibility constraint."

	^Address!

objectConstraint
	"Default bottom compatibility constraint."

	^ActiveObject! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Activity class
	instanceVariableNames: ''!


!Activity class methodsFor: 'process cleanup'!

allCleanUp
	"Provide a reset/cleanup of activity processes of that kind.
	That is, terminate activity processes of all instances of this activity class and of its subclasses."

	"Following expression provides cleanup of all kinds of activities. Takes a little while.
	(On the Mac, move the mouse from time to time!!!!). Print should be reinitialized then."

	"Activity allCleanUp"

	Transcript show: 'CleanUp of ' , self name , ' hierarchy start'.
	self cleanUp.
	self allSubclasses do: [:anActivityClass |
		anActivityClass cleanUp.
		Transcript show: '.'].
	self postCleanUp.	"Some reinitialization may need to be done (to restart aborted activities)."
	Transcript show: ' done!!'; cr!

cleanUp
	"Provide a reset/cleanup of activity processes of this class."
	"See method allCleanUp to also cleanup all subclasses."

	self allInstances do:
		[:anActivity | anActivity terminate]!

gcTest: isCleanUp
	"ActiveObject gcTest: true"		"To show that processes are recovered by gc thanks to cleanup."
	"ActiveObject gcTest: false"	"To crash the system... and therefore prove the need for cleanup."

	"Test of cleanup to allow recovery of terminated processes.
	If cleanup is set (true), it should run infinitely. Stop it eventually by interrupt (^C).
	Otherwise (false), Smalltalk eventually crashes because processes cannot be recovered by gc."

	| n |
	n := 0.
	[true] whileTrue:
		[100 timesRepeat:
			[ActiveObject new active.
			n := n + 1].
		Transcript show: n printString , ' '.
		isCleanUp
			ifTrue: [ActiveObject cleanUp]]!

postCleanUp
	"We need to possibly reset the Print active object whose activity may have been terminated."

	self = Activity
		ifTrue: [Printer initialize]! !
